from __future__ import annotations

import heapq
from abc import ABC, abstractmethod
from collections import deque
from collections.abc import AsyncIterator, Callable
from typing import Generic, Literal, Type, TypeVar

import trio
from attrs import define, field

from aqueue.display import Display


class Item(ABC):
    """An abstract class for items."""

    @abstractmethod
    async def process(self, enqueue: EnqueueFn, progress_display: Display) -> None:
        """
        Do this items work. If any async primitives are to be used in this method, they
        must be compatible with trio.
        """


# type for the enqueue function
EnqueueFn = Callable[[Item], None]


ItemType = TypeVar("ItemType", bound=Item)


@define
class Queue(Generic[ItemType]):
    """
    An unbounded FIFO queue
    """

    _container: deque[ItemType] = field(init=False, factory=deque)
    _unfinished_task_count: int = field(init=False, default=0)

    def _put(self, item: ItemType) -> None:
        self._container.appendleft(item)

    def _get(self) -> ItemType:
        return self._container.pop()

    def put(self, item: ItemType) -> None:
        self._put(item)
        self._unfinished_task_count += 1

    async def get(self) -> ItemType:
        while self.empty():
            await trio.sleep(0)
        return self._get()

    def empty(self) -> bool:
        return not self._container

    def size(self) -> int:
        return len(self._container)

    def task_done(self) -> None:
        self._unfinished_task_count -= 1

    async def join(self) -> None:
        while self._unfinished_task_count != 0:
            await trio.sleep(0)

    async def __aiter__(self) -> AsyncIterator[ItemType]:
        # gather 2 things: a get and a join
        not_changed = object()
        item: object | ItemType = not_changed

        async def join_and_cancel(cancel_scope: trio.CancelScope) -> None:
            await self.join()
            cancel_scope.cancel()

        async def get_and_cancel(cancel_scope: trio.CancelScope) -> None:
            nonlocal item
            item = await self.get()
            cancel_scope.cancel()

        while True:
            async with trio.open_nursery() as nursery:
                nursery.start_soon(join_and_cancel, nursery.cancel_scope)
                nursery.start_soon(get_and_cancel, nursery.cancel_scope)

            if item is not_changed:
                return
            else:
                yield item  # type: ignore

            item = not_changed  # reset for next iter


@define
class Stack(Queue, Generic[ItemType]):
    """
    An unbounded LIFO queue (or stack)
    """

    def _put(self, item: ItemType) -> None:
        self._container.append(item)


@define
class PriorityQueue(Queue, Generic[ItemType]):
    """
    An unbounded priority queue. Items will need to set their own ordering.
    """

    _container: deque[ItemType] = field(init=False, factory=list)

    def _put(self, item: ItemType) -> None:
        heapq.heappush(self._container, item)  # type: ignore

    def _get(self) -> ItemType:
        return heapq.heappop(self._container)  # type: ignore


# names a type of queue for the API
QueueTypeName = Literal["queue", "stack", "priority"]

QUEUE_FACTORY: dict[QueueTypeName, Type[Queue]] = {
    "queue": Queue,
    "stack": Stack,
    "priority": PriorityQueue,
}



from aqueue.display import Display
from aqueue.processing import async_run_queue, run_queue
from aqueue.queue import EnqueueFn, Item, QueueTypeName

__all__ = (
    "Display",
    "EnqueueFn",
    "Item",
    "QueueTypeName",
    "run_queue",
    "async_run_queue",
)



from __future__ import annotations

from collections.abc import Callable, Iterable
from functools import cached_property, wraps
from typing import TYPE_CHECKING

from attrs import define, frozen
from rich.live import Live
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    Progress,
    ProgressColumn,
    SpinnerColumn,
    Task,
    TaskID,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
)
from rich.table import Table

if TYPE_CHECKING:
    # avoid circular reference when just trying to type hint
    from aqueue.queue import Queue


@frozen(kw_only=True)
class DisplaySetuper:
    """Internal display object that organizes away the complex setup of it all"""

    live: Live
    overall_progress: Progress
    overall_progress_task_id: TaskID
    worker_status_progress: Progress
    queue_stats_progress: Progress
    queue_stats_progress_task_id: TaskID

    @classmethod
    def create(
        cls, overall_progress_columns: Iterable[ProgressColumn]
    ) -> DisplaySetuper:
        worker_status_progress = Progress(
            SpinnerColumn(),
            TextColumn("[blue]{task.fields[worker_id]}"),
            TextColumn("[white]{task.description}"),
        )

        queue_stats_progress = Progress(
            SpinnerColumn(),
            TextColumn("[blue]Stack Size"),
            TextColumn("[white]{task.completed} items"),
            BarColumn(complete_style="cyan", finished_style="cyan"),
            TextColumn("[white]{task.total} items max"),
        )
        queue_stats_progress_task_id = queue_stats_progress.add_task("", total=0)

        overall_progress_columns = overall_progress_columns or [
            SpinnerColumn(),
            TextColumn("[blue]{task.description}"),
            TaskProgressColumn(show_speed=True),
            TextColumn("[green]{task.completed}"),
            TimeElapsedColumn(),
            BarColumn(),
        ]
        overall_progress = Progress(*overall_progress_columns)
        overall_progress_task_id = overall_progress.add_task("Overall", total=None)

        table = Table.grid()
        table.add_row(Panel(worker_status_progress, title="Worker Status"))
        table.add_row(Panel(queue_stats_progress, title="Queue Stats"))
        table.add_row(Panel(overall_progress, title="Overall Progress"))

        live = Live(table)

        return DisplaySetuper(
            live=live,
            overall_progress=overall_progress,
            overall_progress_task_id=overall_progress_task_id,
            worker_status_progress=worker_status_progress,
            queue_stats_progress=queue_stats_progress,
            queue_stats_progress_task_id=queue_stats_progress_task_id,
        )

    def create_update_queue_size_progress_fn(self, queue: Queue) -> Callable[..., None]:
        def update_queue_size_progress() -> None:
            new_size = queue.size()
            old_max = (
                self.queue_stats_progress._tasks[
                    self.queue_stats_progress_task_id
                ].total
                or 0
            )

            self.queue_stats_progress.update(
                task_id=self.queue_stats_progress_task_id,
                completed=new_size,
                total=max(old_max, new_size),
            )

        return update_queue_size_progress

    def create_progress_display(
        self, worker_status_progress_task_id: TaskID
    ) -> Display:
        return Display(
            worker=WorkerStatus(
                linked_task=LinkedTask(
                    progress=self.worker_status_progress,
                    task_id=worker_status_progress_task_id,
                ),
            ),
            overall=Overall(
                linked_task=LinkedTask(
                    progress=self.overall_progress,
                    task_id=self.overall_progress_task_id,
                )
            ),
        )


@frozen(kw_only=True, slots=False)
class LinkedTask:
    """
    A progress and a task, linked together. This makes the rich API easier to work with
    because only one thing needs to be passed around, not two.
    """

    progress: Progress
    task_id: TaskID

    @cached_property
    def task(self) -> Task:
        # yeesh, hacky. rich has questionable interface here
        return self.progress._tasks[self.task_id]

    @wraps(Progress.update)
    def update(self, **kwargs) -> None:
        self.progress.update(task_id=self.task_id, **kwargs)


@define(kw_only=True)
class Overall:
    """Represents the "Overall Progress" display."""

    _linked_task: LinkedTask

    @property
    def completed(self) -> float:
        """Get the amount of completed items"""
        return self._linked_task.task.completed

    @completed.setter
    def completed(self, value: float) -> None:
        """Set the amount of completed items"""
        self._linked_task.update(completed=value)

    @property
    def total(self) -> float | None:
        """Get the amount of total items"""
        return self._linked_task.task.total

    @total.setter
    def total(self, value: float | None) -> None:
        """Set the amount of total items"""
        self._linked_task.update(total=value)

    @property
    def total_f(self) -> float:
        """Get the amount of total items. If the total is actually None, return 0."""
        return self._linked_task.task.total or 0

    @total_f.setter
    def total_f(self, value: float) -> None:
        """Set the amount of total items. (Setter paired with total_f.)"""
        self._linked_task.update(total=value)


@define(kw_only=True)
class WorkerStatus:
    """Represents the current worker display."""

    _linked_task: LinkedTask

    @property
    def description(self) -> str:
        """Get the current worker status description"""
        return self._linked_task.task.description

    @description.setter
    def description(self, value: str) -> None:
        """Set the current worker status description"""
        self._linked_task.update(description=value)


@frozen(kw_only=True)
class Display:
    """
    Provides access to queue items to update the display. This is a thin wrapper around
    some rich.progress.Progress methods.
    """

    worker: WorkerStatus
    overall: Overall



from __future__ import annotations

from collections.abc import Callable, Iterable
from functools import partial

import trio
from rich.progress import ProgressColumn

from aqueue.display import Display, DisplaySetuper
from aqueue.queue import QUEUE_FACTORY, Item, Queue, QueueTypeName

_WAIT_MESSAGE = "Waiting for work..."


async def _worker(
    queue: Queue[Item],
    progress_display: Display,
    update_queue_size_progress: Callable[..., None],
    graceful_ctrl_c: bool,
):
    async for item in queue:
        with trio.CancelScope(shield=graceful_ctrl_c):
            await item.process(queue.put, progress_display)
        progress_display.worker.description = _WAIT_MESSAGE
        update_queue_size_progress()
        queue.task_done()

    progress_display.worker.description = "Done"


async def async_run_queue(
    *,
    queue_type_name: QueueTypeName,
    overall_progress_columns: Iterable[ProgressColumn] | None = None,
    initial_items: Iterable[Item] | None = None,
    num_workers: int = 10,
    graceful_ctrl_c: bool = True,
) -> None:
    """
    An asynchronous version of `run_queue`. This method must be run in a Trio event
    loop.
    """
    queue = QUEUE_FACTORY[queue_type_name]()

    display = DisplaySetuper.create(
        overall_progress_columns=overall_progress_columns or []
    )
    update_queue_size_progress = display.create_update_queue_size_progress_fn(queue)

    display.live.start()

    async with trio.open_nursery() as nursery:
        for item in initial_items or []:
            queue.put(item)

        for i in range(num_workers):
            worker_status_progress_task_id = display.worker_status_progress.add_task(
                _WAIT_MESSAGE, worker_id=f"#{i}"
            )
            progress_display = display.create_progress_display(
                worker_status_progress_task_id
            )

            nursery.start_soon(
                partial(
                    _worker,
                    queue=queue,
                    progress_display=progress_display,
                    update_queue_size_progress=update_queue_size_progress,
                    graceful_ctrl_c=graceful_ctrl_c,
                ),
                name=f"worker #{i}",
            )

    display.live.stop()


def run_queue(
    *,
    num_workers: int = 10,
    initial_items: Iterable[Item] | None = None,
    queue_type_name: QueueTypeName = "queue",
    overall_progress_columns: Iterable[ProgressColumn] | None = None,
    graceful_ctrl_c: bool = True,
) -> None:
    """
    Process all items in initial items (and any subsequent items they produce) and
    display a terminal visualization of it.

    - `num_workers` specifies how many workers will be running concurrently
    - `initial_items` is an iterable that seeds the queue. This is where the top-level
      item should go that produces more items. (Note that any subsequent item can also
      produce items.)
    - `queue_type_name` can be either of:
      - `queue` for first-in-first-out processing
      - `stack` for last-in-first-out processing
      - `priority` for priority-based processing. In this case, item objects should be
        orderable. Processing will occur in *ascending* priority (smallest first).
    - `overall_progress_columns` is an iterable of columns for the "Overall Progress"
      panel. These must be `rich.progress.ProgressColumn` objects. See
      https://rich.readthedocs.io/en/stable/progress.html#columns.
    - `graceful_ctrl_c` specifies whether pressing Ctrl-C will stop things abruptly
      (False) or wait until all the currently worked-on items are finished first (True,
      the default).
    """

    trio.run(
        partial(
            async_run_queue,
            queue_type_name=queue_type_name,
            initial_items=initial_items,
            overall_progress_columns=overall_progress_columns,
            num_workers=num_workers,
            graceful_ctrl_c=graceful_ctrl_c,
        ),
        restrict_keyboard_interrupt_to_checkpoints=False,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['aqueue']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=22.1.0,<23.0.0', 'rich>=12.6.0,<13.0.0', 'trio>=0.22.0,<0.23.0']

setup_kwargs = {
    'name': 'aqueue',
    'version': '0.2.0',
    'description': 'An async queue with live progress display',
    'long_description': '# aqueue\n\n![demo](docs/demo.gif)\n\nAn async queue with live progress display. Good for running and visualizing tree-like I/O-bound\nprocessing jobs, such as website scrapes.\n\n## Example\n\n```python\nimport random\n\nimport trio\n\nfrom aqueue import EnqueueFn, Display, run_queue, Item\n\n\nclass RootItem(Item):\n    async def process(self, enqueue: EnqueueFn, display: Display) -> None:\n        num_children = 3\n        display.overall.total = num_children\n        display.worker.description = "Making child items"\n\n        for _ in range(num_children):\n            # simulate doing work and creating more items\n            await trio.sleep(random.random())\n            enqueue(ChildItem())\n\n\nclass ChildItem(Item):\n    async def process(self, enqueue: EnqueueFn, display: Display) -> None:\n        display.worker.description = "Doing work..."\n\n        # Simulate doing work\n        await trio.sleep(random.random())\n\n        display.overall.completed += 1\n\n\ndef main() -> None:\n    run_queue(\n        initial_items=[RootItem()],\n        num_workers=2,\n    )\n\n\nif __name__ == "__main__":\n    main()\n\n```\n\n## Usage Notes\n\nThere\'s two things you need to do to use aqueue:\n\n1. Write your [Item](#items) classes\n2. [Start your queue](#starting-your-queue) with one of those items\n\n### Items\n\nItems are your units of work. They can represent whatever you\'d like, such as parts of a website\nthat you\'re trying to scrape: an item for the index page, for subpages, for images, etc.\n\nEach item should be an instance of a class that defines an async `progress` method. As arguments, it\nshould accept two positional arguments:\n\n1. a `aqueue.EnqueueFn` that caan be called to enqueue more work. That type is simply an alias for\n   `Callable[[Item], None]`.\n2. a `aqueue.Display` object that gives you control of the terminal display:\n\n```python\nimport aqueue\n\nclass MyItem(aqueue.Item):\n    async def process(self, enqueue: aqueue.EnqueueFn, display: aqueue.Display) -> None:\n        # make an HTTP request, parse it, etc\n        print(\'My item is processing!\')\n\n        # when you discover more items you want to process, enqueue them:\n        enqueue(AnotherItem())\n\nclass AnotherItem(aqueue.Item):\n    async def process(self, enqueue: aqueue.EnqueueFn, display: aqueue.Display) -> None:\n        print(\'Another item is processing!\')\n```\n\nAs a rule of thumb, you should make a new item class whenever you notice a one-to-many relationship.\nFor example, this _one_ page has _many_ images I want to download.\n\nNote: `process` is async, but because this library uses\n[Trio](https://trio.readthedocs.io/en/stable/index.html) under the hood, you may only use\nTrio-compatible primitives inside `process`. For example, use `trio.sleep`, not `asyncio.sleep`.\nTODO: consider [AnyIO](https://anyio.readthedocs.io/en/stable/) to avoid this problem?\n\nDisclaimer: aqueue, or any asynchronous framework, is only going to be helpful if you\'re performing\nwork is I/O-bound.\n\n### Starting your Queue\n\nThen, start your queue with an initial item(s) to kick things off.\n\n```python\naqueue.run_queue(\n    initial_items=[MyItem()],\n    num_workers=2,\n)\n```\n\n#### Queue type\n\nBy default, the queue is actually ...a queue -- that is to say that items are processed\nfirst-in-first-out. Here are all the types you can specify with the `queue_type_name` argument.\n\n- `queue` - first-in-first-out processing, or breadth-first.\n- `stack` - last-in-first-out processing, or depth-first. This one is recommended for website\n  scraping because it yields items fast (versus `queue` that tries to look at all the intermediate\n  pages first).\n- `priority` - priority queue processing. In this case, your item objects should be orderable (with\n  `__lt__`, etc). **Lesser objects will be processed first**, because this code uses a minheap.\n\n#### Number of workers\n\nYou can specify the number of workers you\'d like to be processing your items with the `num_workers`\nargument.\n\n#### Ctrl-C\n\nIf you decide you want to stop your queue processing, press Ctrl-C.\n\nIf you\'ve set the `graceful_ctrl_c` to False, this will stop the program immediately. If True, the\ndefault, aqueue will wait for the items currently being worked on to complete (without taking any\nadditional items), and _then_ stop. Put another way, the choice is between responsiveness and\nresource consistency.\n\n#### Setting the look of the panels\n\nCurrently, only support for configuring the "Overall Progress" panel is supported. By default, the\npanel is very simple. If you want to customize it, provide an iterable of\n`rich.progress.ProgressColumn` objects to the `overall_progress_columns` argument. See\n<https://rich.readthedocs.io/en/stable/progress.html> for more information. (Note that rich provides\nall the nice terminal visualizations for aqueue!)\n\n### Updating the display\n\nAs mentioned, each `process` method gets called with an `aqueue.Display` object. The display has two\nproperties:\n\n- `worker`, which lets you update the description of the worker who\'s currently processing this\n  item. `display.worker.description` is the getter/setter for that.\n- `overall`, which lets you access things in "Overall Progress" terminal panel.\n  `display.overall.completed` is a getter/setter for the number of completed things,\n  `display.overall.total` for the total number of things (or None), and `display.overall.total_f`\n  for the total number of things or 0.\n\nThese panels are just an informational display for humans. They don\'t know about the queue churning\nthrough items of work. Therefore, you must decide what things you want to keep track of, and often,\nyou won\'t be able to determine the complete number of things at the beginning. You\'ll need to do\nsome intermediate processing and increment it slowly as more work is discovered. For example, if you\nwant to keep track of images found and downloaded, you often won\'t be able to do that until you are\nsearching deeper into the website.\n\n### Sharing state\n\nOften, its beneficial to share state between the items. Using the website scrape example again, you\nmay want to keep track of the URLs you\'ve visited so you don\'t scrape them twice.\n\nIf this is needed, simply keep a global set/dict/list and store a key for the item. For example, a\nURL string may be a good key.\n\nIf you don\'t want to or can\'t use a global variable, consider a\n[`ContextVar`](https://docs.python.org/3/library/contextvars.html).\n\n### Persisting state\n\nDuring development, its probably likely that your program will crash after doing some work. For\nexample, maybe your HTTP request timed out or you had a bug in your HTML parsing.\n\nIt\'s a shame to lose that work that\'s been done. So, if you\'re looking for a really handy way to\npersist state across runs, check out the built-in\n[`shelve`](https://docs.python.org/3/library/shelve.html) module. It\'s like a dict that\nautomatically saves to a file each time you set a key in it.\n\n### Other cool things\n\nThis library is fully docstringed and type-hinted 🥳\n\n## Installation\n\n```shell\npip install "git+https://github.com/t-mart/aqueue"\n```\n',
    'author': 'Tim Martin',
    'author_email': 'tim@timmart.in',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
